// (C) Copyright 2002-2010 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// AssocArrayDemo.cpp illustrates use of the Associative Array API.
//
// Products: AutoCAD (and verticals) 2012 and higher
//
// Written by Philippe Leefsma - 2010 - Developer Technical Services
//     AutoCAD 2013+ -  
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#include "StdAfx.h"
#include "AssocArrayDemo.h"
#include "AcDbAssocManager.h"
#include "AcDbAssocArrayActionBody.h"
#include "AcDbAssocArrayPathParameters.h"
#include "AcDbAssocArrayPolarParameters.h"
#include "AcDbAssocArrayModifyActionBody.h"
#include "AcDbAssocArrayRectangularParameters.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Acad::ErrorStatus postToDatabase (AcDbEntity* pEntity,  
  AcDbDatabase* pDb = acdbHostApplicationServices()->workingDatabase()) 
{
  AcDbBlockTablePointer blkTable(pDb, AcDb::kForRead);

  if(blkTable.openStatus() != Acad::eOk)
    return blkTable.openStatus();

  AcDbBlockTableRecordPointer btr(ACDB_MODEL_SPACE, pDb, AcDb::kForWrite);

  if(btr.openStatus() != Acad::eOk)
    return btr.openStatus();

  return btr->appendAcDbEntity(pEntity);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// :    
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Acad::ErrorStatus AssocArrayDemo::createRectangularArray()
{
  Acad::ErrorStatus es;

  double columnSpacing = 5;
  double rowSpacing = 5;
  double levelSpacing = 1;
  int columnCount = 5;
  int rowCount = 4;
  int levelCount = 1;
  double rowElevation = 1;
  double axesAngle = 90;

  //     
  AcDbAssocArrayRectangularParameters* rectParams = new AcDbAssocArrayRectangularParameters(columnSpacing,
    rowSpacing,
    levelSpacing,
    columnCount,
    rowCount,
    levelCount,
    rowElevation,
    axesAngle);

  //   ( )  
  AcDbObjectIdArray profileEntityIdArray;

  AcDbCircle* pCircle = new AcDbCircle(AcGePoint3d(0, 5, 0), AcGeVector3d(0, 0, 1), 1.5);

  pCircle->setDatabaseDefaults();

  if( (es = postToDatabase(pCircle)) != Acad::eOk )
  {
    delete pCircle;
    return es;
  }

  profileEntityIdArray.append(pCircle->objectId());

  pCircle->close();

  //   
  AcDbVertexRef basePoint(AcGePoint3d(0, 0, 0));

  //  :     ,    . 
  //     AcDbAssocArrayActionBody::createInstance,   arrayId  actionBodyId
  // .         ObjectARX C++  AutoCAD .NET API
  AcDbObjectId arrayId;
  AcDbObjectId actionBodyId;	

  if( (es = AcDbAssocArrayActionBody::createInstance(profileEntityIdArray, basePoint, rectParams, arrayId, actionBodyId)) != Acad::eOk)
    return es;


  //       .       AcDbEntity  arrayId.
  //   ,      .
  AcDbEntity* pArray = NULL;

  if( (es = acdbOpenAcDbObject((AcDbObject*&)pArray, arrayId, AcDb::kForWrite)) != Acad::eOk)
    return es;

  AcGeMatrix3d Transfo;
  Transfo.setToTranslation(AcGeVector3d(-25,0,0));

  es = pArray->transformBy(Transfo);
  es = pArray->close();

  //    ,       
  AcDbAssocManager::evaluateTopLevelNetwork(arrayId.database());
  acedUpdateDisplay();	

  return es;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// :   : /    
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Acad::ErrorStatus AssocArrayDemo::modifyArray(AcDbObjectId& arrayId)
{
  Acad::ErrorStatus es;

  AcDbObjectPointer <AcDbEntity> pEntity(arrayId, AcDb::kForRead);

  if((es = pEntity.openStatus()) != Acad::eOk)
    return es;

  //     "isAssociativeArray"        
  if(!AcDbAssocArrayActionBody::isAssociativeArray(pEntity))
    return Acad::eInvalidInput;

  //   AcDbAssocArrayActionBody   
  AcDbObjectId actionBodyId = AcDbAssocArrayActionBody::getControllingActionBody(pEntity, NULL);

  AcDbAssocArrayActionBody* pArrayActionBody = NULL;

  if( (es = acdbOpenAcDbObject((AcDbObject*&)pArrayActionBody, actionBodyId, AcDb::kForWrite)) != Acad::eOk)
    return es;


  AcArray<AcDbItemLocator> indices;

  //  ,  ,  
  AcDbItemLocator deletedItem(2, 2, 0);

  indices.append(indices);

  //    
  if((es = pArrayActionBody->deleteItem(deletedItem, true)) != Acad::eOk)
  {
    pArrayActionBody->close();
    return es;
  }

  pArrayActionBody->close();

  //  ,  ,  
  AcDbItemLocator replacedItem(0, 0, 0);

  indices.removeAll();
  indices.append(replacedItem);


  //   ( )      
  AcDbObjectIdArray substEntities;

  AcDbPolyline* pLine = new AcDbPolyline();

  pLine->setDatabaseDefaults();
  pLine->addVertexAt(pLine->numVerts(), AcGePoint2d(-1, 4));
  pLine->addVertexAt(pLine->numVerts(), AcGePoint2d(-1, 6));
  pLine->addVertexAt(pLine->numVerts(), AcGePoint2d(1, 6));
  pLine->addVertexAt(pLine->numVerts(), AcGePoint2d(1, 4));
  pLine->setClosed(true);
  pLine->setColorIndex(1);

  if( (es = postToDatabase(pLine)) != Acad::eOk )
  {
    delete pLine;
    return es;
  }

  substEntities.append(pLine->objectId());

  pLine->close();


  AcDbVertexRef basePoint(AcGePoint3d(0,0,0));

  AcDbObjectId modifyActionBodyId;

  //    
  if( (es = AcDbAssocArrayModifyActionBody::createInstance(arrayId, indices, substEntities, basePoint, modifyActionBodyId)) != Acad::eOk)
    return es;

  //    ,       
  AcDbAssocManager::evaluateTopLevelNetwork(arrayId.database());
  acedUpdateDisplay();	

  return es;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// :    (     )
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Acad::ErrorStatus AssocArrayDemo::explodeArray(AcDbObjectId& arrayId)
{
  Acad::ErrorStatus es;

  AcDbObjectPointer <AcDbEntity> pEntity(arrayId, AcDb::kForRead);

  if((es = pEntity.openStatus()) != Acad::eOk)
    return es;

  //     "isAssociativeArray"        
  if(!AcDbAssocArrayActionBody::isAssociativeArray(pEntity))
    return Acad::eInvalidInput;

  AcDbObjectIdArray newIds;

  if((es = AcDbAssocArrayActionBody::explode(pEntity, newIds)) != Acad::eOk)
    return es;

  AcDbAssocManager::evaluateTopLevelNetwork(acdbCurDwg());
  acedUpdateDisplay();	

  return es;
}

//    
std::wstring objIdToEnameStr(AcDbObjectId& id);
std::wstring directionToStr(AcDbAssocArrayPolarParameters::Direction direction);
std::wstring methodToStr(AcDbAssocArrayPathParameters::Method method);

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// :      
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Acad::ErrorStatus AssocArrayDemo::infoArray(AcDbObjectId& arrayId)
{
  Acad::ErrorStatus es;

  AcDbObjectPointer <AcDbEntity> pEntity(arrayId, AcDb::kForRead);

  if((es = pEntity.openStatus()) != Acad::eOk)
    return es;

  //     "isAssociativeArray"        
  if(!AcDbAssocArrayActionBody::isAssociativeArray(pEntity))
    return Acad::eInvalidInput;


  acutPrintf(L"\n    ----     ---- ");

  //  AcDbAssocArrayActionBody   
  AcDbObjectId actionBodyId = AcDbAssocArrayActionBody::getControllingActionBody(pEntity, NULL);

  acutPrintf(L"\n -  AcDbAssocArrayActionBody: %s", objIdToEnameStr(actionBodyId).c_str());

  AcDbAssocArrayActionBody* pArrayActionBody = NULL;

  if( (es = acdbOpenAcDbObject((AcDbObject*&)pArrayActionBody, actionBodyId, AcDb::kForWrite)) != Acad::eOk)
    return es;


  acutPrintf(L"\n - :");

  //       
  const AcDbAssocArrayParameters* pParameters = pArrayActionBody->parameters();

  AcDbAssocArrayCommonParameters* pComonParameters = AcDbAssocArrayCommonParameters::cast(pParameters);

  AcString expression; 

  int levelCount = pComonParameters->levelCount(expression);
  acutPrintf(L"\n  .  = %d (: \"%s\")", levelCount, (expression.length() != 0 ? expression.kwszPtr() : L""));

  double levelSpacing = pComonParameters->levelSpacing(expression);
  acutPrintf(L"\n  .    = %f (: \"%s\")", levelSpacing, (expression.length() != 0 ? expression.kwszPtr() : L""));

  int rowCount = pComonParameters->rowCount(expression);
  acutPrintf(L"\n  .  = %d (: \"%s\")", rowCount, (expression.length() != 0 ? expression.kwszPtr() : L""));

  double rowElevation = pComonParameters->rowElevation(expression);
  acutPrintf(L"\n  .    = %f (: \"%s\")", rowElevation, (expression.length() != 0 ? expression.kwszPtr() : L""));

  double rowSpacing = pComonParameters->rowSpacing(expression);
  acutPrintf(L"\n  .    = %f (: \"%s\")", rowSpacing, (expression.length() != 0 ? expression.kwszPtr() : L""));

  //   
  if(pComonParameters->isKindOf(AcDbAssocArrayRectangularParameters::desc()))
  {
    acutPrintf(L"\n -   :");

    AcDbAssocArrayRectangularParameters* pRectParameters = AcDbAssocArrayRectangularParameters::cast(pComonParameters);

    int columnCount = pRectParameters->columnCount(expression);
    acutPrintf(L"\n  .   = %d (: \"%s\")", columnCount, (expression.length() != 0 ? expression.kwszPtr() : L""));

    double columnSpacing = pRectParameters->columnSpacing(expression);
    acutPrintf(L"\n  .    = %f (: \"%s\")", columnSpacing, (expression.length() != 0 ? expression.kwszPtr() : L""));

    double axesAngle = pRectParameters->axesAngle(expression);
    acutPrintf(L"\n  .   = %f (: \"%s\")", axesAngle, (expression.length() != 0 ? expression.kwszPtr() : L""));

    AcGeVector3d axisDirection = pRectParameters->axisDirection();
    acutPrintf(L"\n  .   = [%f, %f, %f]", axisDirection.x, axisDirection.y, axisDirection.z);
  }

  //   
  if(pComonParameters->isKindOf(AcDbAssocArrayPolarParameters::desc()))
  {
    acutPrintf(L"\n -   :");

    AcDbAssocArrayPolarParameters* pPolarParameters = AcDbAssocArrayPolarParameters::cast(pComonParameters);

    int itemCount = pPolarParameters->itemCount(expression);
    acutPrintf(L"\n  .   = %d (: \"%s\")", itemCount, (expression.length() != 0 ? expression.kwszPtr() : L""));

    double angleBetweenItems = pPolarParameters->angleBetweenItems(expression);
    acutPrintf(L"\n  .    = %f (: \"%s\")", angleBetweenItems, (expression.length() != 0 ? expression.kwszPtr() : L""));

    double fillAngle = pPolarParameters->fillAngle(expression);
    acutPrintf(L"\n  .   = %f (: \"%s\")", fillAngle, (expression.length() != 0 ? expression.kwszPtr() : L""));
    AcString evaluatorId;
    double startAngle = pPolarParameters->startAngle(expression, evaluatorId);
    acutPrintf(L"\n  .   = %f (: \"%s\")", startAngle, (expression.length() != 0 ? expression.kwszPtr() : L""));

    AcDbAssocArrayPolarParameters::Direction direction = pPolarParameters->direction();
    acutPrintf(L"\n  .  = %s", directionToStr(direction).c_str());

    bool rotateItems = pPolarParameters->rotateItems();
    acutPrintf(L"\n  .   = %s", (rotateItems ? L"" : L""));

    double radius = pPolarParameters->radius(expression);
    acutPrintf(L"\n  .  = %f (: \"%s\")", radius, (expression.length() != 0 ? expression.kwszPtr() : L""));
  }

  //   
  if(pComonParameters->isKindOf(AcDbAssocArrayPathParameters::desc()))
  {
    acutPrintf(L"\n -   :");

    AcDbAssocArrayPathParameters* pPathParameters = AcDbAssocArrayPathParameters::cast(pComonParameters);

    int itemCount = pPathParameters->itemCount(expression);
    acutPrintf(L"\n  .   = %d (: \"%s\")", itemCount, (expression.length() != 0 ? expression.kwszPtr() : L""));

    double itemSpacing = pPathParameters->itemSpacing(expression);
    acutPrintf(L"\n  .    = %f (: \"%s\")", itemSpacing, (expression.length() != 0 ? expression.kwszPtr() : L""));

    AcDbEdgeRef path = pPathParameters->path();
    AcDbEntity* pathEnt = path.createEntity();
    acutPrintf(L"\n  .   = %s ", pathEnt->isA()->name());
    delete pathEnt;

    AcDbAssocArrayPathParameters::Method method = pPathParameters->method();
    acutPrintf(L"\n  .  = %s", methodToStr(method).c_str());

    bool alignItems = pPathParameters->alignItems();
    acutPrintf(L"\n  .   = %s", (alignItems ? L"" : L""));
    AcString evaluatorId;
    double startOffset = pPathParameters->startOffset(expression,evaluatorId);
    acutPrintf(L"\n  .   = %f (: \"%s\")", startOffset, (expression.length() != 0 ? expression.kwszPtr() : L""));

    double endOffset = pPathParameters->endOffset(expression,evaluatorId);
    acutPrintf(L"\n  .   = %f (: \"%s\")", endOffset, (expression.length() != 0 ? expression.kwszPtr() : L""));
  }


  //     
  AcDbObjectId btrId = pArrayActionBody->getArraySourceBTR();
  acutPrintf(L"\n -  BlockTableRecord: %s", objIdToEnameStr(btrId).c_str());


  acutPrintf(L"\n -  :");

  //   
  AcDbObjectIdArray sourceEntityIds = pArrayActionBody->getSourceEntities();

  for(int i=0; i<sourceEntityIds.length(); ++i)
  {
    AcDbObjectPointer <AcDbEntity> pSourceEntity(sourceEntityIds[i], AcDb::kForRead);

    if( (es = pSourceEntity.openStatus()) != Acad::eOk)
      continue;

    const ACHAR* name = pSourceEntity->isA()->name();

    acutPrintf(L"\n  . %s", name);
  }


  acutPrintf(L"\n -   :");

  // ,    vertexRef        ,
  //       ( )  BTR
  AcDbVertexRef vertexRef;
  AcGePoint3d position;

  es = pArrayActionBody->getSourceBasePoint(vertexRef, position);

  acutPrintf(L"\n  . VertexRef: [%f, %f, %f]", vertexRef.point().x, vertexRef.point().y, vertexRef.point().z);
  acutPrintf(L"\n  . Position: [%f, %f, %f]", position.x, position.y, position.z);


  //           -   
  // ActionBody,  ,   .
  acutPrintf(L"\n -  :");

  AcArray<AcDbItemLocator> indices;

  bool skipErased = true;

  pArrayActionBody->getItems(indices, skipErased);

  for(int i=0; i<indices.length(); ++i)
  {
    AcDbItemLocator itemLocator = indices[i];

    AcDbFullSubentPath path;

    const AcDbAssocArrayItem* arrayItem = pArrayActionBody->getItemAt(itemLocator, path);

    // ,   ActionBody   ,   ActionBody  
    //  ,  .
    if(!pArrayActionBody->controlsItem(*arrayItem))
    {
      int itemIndex = itemLocator[AcDbItemLocator::kItemIndex];

      acutPrintf(L"\n  .  : %d", itemIndex);

      //   ActionBody
      AcDbObjectId modifyActionBodyId = AcDbAssocArrayActionBody::getControllingActionBody(pEntity, &itemLocator);

      acutPrintf(L"\n  . AcDbAssocArrayModifyActionBody: %s", objIdToEnameStr(modifyActionBodyId).c_str());

      AcDbAssocArrayModifyActionBody* pArrayModifyActionBody = NULL;

      if( (es = acdbOpenAcDbObject((AcDbObject*&)pArrayModifyActionBody, modifyActionBodyId, AcDb::kForRead)) == Acad::eOk)
      {
        AcArray<AcDbItemLocator> overridenItems;
        es = pArrayModifyActionBody->getOverridenItems(overridenItems);

        acutPrintf(L"\n  .   : %i", overridenItems.length());

        pArrayModifyActionBody->close();
      }
    }
  }

  pArrayActionBody->close();

  return es;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// :  ( , ObjectId  , enum  , ...)
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
std::wstring StringToWString(const std::string& s)
{
  std::wstring temp(s.length(),L' ');
  std::copy(s.begin(), s.end(), temp.begin());
  return temp;
}

std::wstring numericToWString(const Adesk::IntDbId& in) 
{
  std::ostringstream ost(std::ostringstream::out);
  ost << in;
  return StringToWString(ost.str());
}

std::wstring objIdToEnameStr(AcDbObjectId& id)
{
  return numericToWString(id.asOldId());
}

std::wstring directionToStr(AcDbAssocArrayPolarParameters::Direction direction)
{
  switch(direction)
  {
  case AcDbAssocArrayPolarParameters::kClockwise:
    return L" ";

  case AcDbAssocArrayPolarParameters::kCounterClockwise:
    return L" ";

  default:
    return L" ";
  }
}

std::wstring methodToStr(AcDbAssocArrayPathParameters::Method method)
{
  switch(method)
  {
  case AcDbAssocArrayPathParameters::kDivide:
    return L"";

  case AcDbAssocArrayPathParameters::kMeasure:
    return L"";

  default:
    return L" ";
  }
}