﻿// (C) Copyright 2002-20010 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// AcHelper is a utility class for the Associative Array Sample
//
// Products: AutoCAD (and verticals) 2011 and higher
//
// Written by Philippe Leefsma - 2010 - Developer Technical Services
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.Text;

using Autodesk.AutoCAD;
using Autodesk.AutoCAD.Runtime;
using Autodesk.AutoCAD.EditorInput;
using Autodesk.AutoCAD.DatabaseServices;
using Autodesk.AutoCAD.ApplicationServices;

[assembly: CommandClass(typeof(AssocArrayAPI.AcHelper))]

namespace AssocArrayAPI
{
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Helper Class
    //
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    public class AcHelper
    {
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Добавление примитива в пространство модели указанной базы данных
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static ObjectId PostToDatabase(Database db, Entity entity)
        {
            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                BlockTable bt = Tx.GetObject(db.BlockTableId, OpenMode.ForRead) as BlockTable;

                BlockTableRecord model = Tx.GetObject(bt[BlockTableRecord.ModelSpace], OpenMode.ForWrite) as BlockTableRecord;

                ObjectId newId = model.AppendEntity(entity);

                Tx.AddNewlyCreatedDBObject(entity, true);

                Tx.Commit();

                return newId;
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Добавление примитива в пространство модели текущей базы данных
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static ObjectId PostToDatabase(Entity entity)
        {
            Database db = Application.DocumentManager.MdiActiveDocument.Database;

            return PostToDatabase(db, entity);
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Получение граней субпримитивов из выбранных при помощи протокола AssocPersSubentityIdPE
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static FullSubentityPath[] GetEdgeSubentities(Entity entity)
        {
            ObjectId[] entId = new ObjectId[] { entity.ObjectId };

            IntPtr pSubentityIdPE = entity.QueryX(AssocPersSubentityIdPE.GetClass(typeof(AssocPersSubentityIdPE)));

            if (pSubentityIdPE == IntPtr.Zero)
                return null;

            AssocPersSubentityIdPE subentityIdPE = AssocPersSubentityIdPE.Create(pSubentityIdPE, false)
                as AssocPersSubentityIdPE;

            SubentityId[] edgeSubentityIds = subentityIdPE.GetAllSubentities(entity, SubentityType.Edge);

            List<FullSubentityPath> edges = new List<FullSubentityPath>();

            foreach (SubentityId s in edgeSubentityIds)
            {
                edges.Add(new FullSubentityPath(entId, s));
            }

            return edges.ToArray();
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Создание PathRef из FullSubentityPath
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static PathRef CreatePathRef(FullSubentityPath subentPath)
        {
            EdgeRef[] edges = new EdgeRef[] { new EdgeRef(subentPath) };

            PathRef path = new PathRef(edges);

            return path;
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Создание PathRef из FullSubentityPath[]
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static PathRef CreatePathRef(FullSubentityPath[] subentPathArray)
        {
            List<EdgeRef> edges = new List<EdgeRef>();

            foreach (FullSubentityPath subentPath in subentPathArray)
            {
                edges.Add(new EdgeRef(subentPath));
            }

            PathRef path = new PathRef(edges.ToArray());

            return path;
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Создание Profile3d из FullSubentityPath
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static Profile3d CreateProfile3d(FullSubentityPath subentPath)
        {
            EdgeRef[] edges = new EdgeRef[] { new EdgeRef(subentPath) };

            Profile3d profile = new Profile3d(new PathRef(edges));

            return profile;
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Helper method: Создание Profile3d из массива FullSubentityPath
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static Profile3d CreateProfile3d(FullSubentityPath[] subentPathArray)
        {
            PathRef path = CreatePathRef(subentPathArray);

            Profile3d profile = new Profile3d(path);

            return profile;
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Добавление новой переменной в ассоциативную сеть или модификация существующей
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static ObjectId AddOrModifVariable(string varName, string varExpression)
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;

            ObjectId varid = ObjectId.Null;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                ObjectId networkId = AssocNetwork.GetInstanceFromObject(db.CurrentSpaceId, true, true, "ACAD_ASSOCNETWORK");
                AssocNetwork network = Tx.GetObject(networkId, OpenMode.ForWrite) as AssocNetwork;

                ObjectIdCollection actionIds = network.GetActions;

                for (int i = 0; i < actionIds.Count; ++i)
                {
                    try
                    {
                        AssocVariable var1 = Tx.GetObject(actionIds[i], OpenMode.ForWrite) as AssocVariable;

                        if (var1 != null)
                        {
                            if (var1.Name == varName)
                            {
                                String errMsg = "";
                                var1.SetExpression(varExpression, "", true, true, ref errMsg, false);

                                ResultBuffer rb = new ResultBuffer();
                                errMsg = var1.EvaluateExpression(ref rb);
                                var1.Value = rb;

                                Tx.Commit();

                                return var1.ObjectId;
                            }
                        }
                    }
                    catch
                    {

                    }
                }

                AssocVariable var = new AssocVariable();
                varid = network.Database.AddDBObject(var);
                network.AddAction(var.ObjectId, true);
                Tx.AddNewlyCreatedDBObject(var, true);
                Tx.Commit();
            }

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                AssocVariable var = (AssocVariable)Tx.GetObject(varid, OpenMode.ForWrite);
                var.SetName(varName, true);
                var.Value = new ResultBuffer(new TypedValue(1001, 5));
                String errMsg = "";
                var.SetExpression(varExpression, "", true, true, ref errMsg, false);
                ResultBuffer rb = new ResultBuffer();
                errMsg = var.EvaluateExpression(ref rb);
                var.Value = rb;
                Tx.Commit();
            }

            return varid;
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Добавление / Модификация ассоциативной переменной
        //
        // Если переменная уже существует, то она обновится выражением введенным в командной строке
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("AddOrModifyAssocVar")]
        public void AddOrModifyAssocVar()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            PromptResult prVar = ed.GetString("\nВведите имя переменной: ");

            if (prVar.Status != PromptStatus.OK) return;

            PromptResult prExpr = ed.GetString("\nВыедите выражение: ");

            if (prExpr.Status != PromptStatus.OK) return;

            ObjectId varid = AddOrModifVariable(prVar.StringResult, prExpr.StringResult);

            ed.WriteMessage("\nVariable Id: " + varid.ToString());
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Печать списка существующих ассоциативных переменных
        //
        // Последовательно проходим по сети текущего пространства и печатаем
        // в командной строке существующие переменные и их выражения
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        [CommandMethod("ListAssocVars")]
        public void ListAssocVars()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            ListAssocVariables(db);
        }

        public void ListAssocVariables(Database db)
        {
            Editor ed = Application.DocumentManager.MdiActiveDocument.Editor;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                ObjectId networkId = AssocNetwork.GetInstanceFromObject(db.CurrentSpaceId, true, true, "ACAD_ASSOCNETWORK");

                AssocNetwork network = Tx.GetObject(networkId, OpenMode.ForRead) as AssocNetwork;

                foreach (ObjectId actionId in network.GetActions)
                {
                    DBObject obj = Tx.GetObject(actionId, OpenMode.ForRead);

                    if (obj is AssocVariable)
                    {
                        AssocVariable assocVar = obj as AssocVariable;
                        ed.WriteMessage("\n - Ассоциативная переменная: " + assocVar.Name + " = " + assocVar.Expression);
                    }
                }
            }
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Использование: 
        //       Запрашиваем у пользователя выбор субпримитивов.
        //       Они все имеют свои ObjectIds и свои пути к субпримитивам
        //       (возвращаемый функцией HighlightSubEntity())
        //       и вся эта коллекция возвращается в вызывающую программу.
        //
        //       Исходник: 
        //       http://through-the-interface.typepad.com/through_the_interface/2010/08/changing-the-layer-or-color-of-a-nested-entity-using-net.html
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        public static PromptStatus SelectNestedEntities(string prompt, out ObjectIdCollection ids, out FullSubentityPath[] paths)
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            using (Transaction Tx = db.TransactionManager.StartTransaction())
            {
                ids = new ObjectIdCollection();

                List<FullSubentityPath> pathsList = new List<FullSubentityPath>();

                PromptNestedEntityOptions pneo = new PromptNestedEntityOptions(prompt);
                pneo.AllowNone = true;

                // Цикл до Отказа или завершения
                while (true)
                {
                    PromptNestedEntityResult rs = ed.GetNestedEntity(pneo);

                    // Выход если отказались
                    if (rs.Status != PromptStatus.OK)
                    {
                        paths = pathsList.ToArray();

                        UnhighlightSubEntities(paths);

                        return (rs.Status != PromptStatus.None ? rs.Status : PromptStatus.OK);
                    }

                    ids.Add(rs.ObjectId);

                    FullSubentityPath path = HighlightSubEntity(rs);

                    if (path != FullSubentityPath.Null)
                        pathsList.Add(path);
                }
            }
        }

        public static PromptStatus SelectNestedEntities(string prompt, out FullSubentityPath[] paths)
        {
            ObjectIdCollection ids = new ObjectIdCollection();

            return SelectNestedEntities(prompt, out ids, out paths);
        }

        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        // Использование: Подсветка субпримитивов на основе информации о выбранных
        //
        //      Source: 
        //      http://through-the-interface.typepad.com/through_the_interface/2010/08/changing-the-layer-or-color-of-a-nested-entity-using-net.html
        //
        /////////////////////////////////////////////////////////////////////////////////////////////////////////
        private static FullSubentityPath HighlightSubEntity(PromptNestedEntityResult rs)
        {
            // Извлекаем соответствующую информацию из выбранного объекта
            ObjectId selId = rs.ObjectId;

            List<ObjectId> objIds = new List<ObjectId>(rs.GetContainers());

            // Обращаем список IDs
            objIds.Reverse();

            // Добавляем выбранный примитив
            objIds.Add(selId);

            // Извлекаем путь к субпримитиву из примитива
            SubentityId subEnt = new SubentityId(SubentityType.Null, System.IntPtr.Zero);

            FullSubentityPath path = new FullSubentityPath(objIds.ToArray(), subEnt);

            // Открываем самый внешний примитив используя текущую транзакцию
            Entity ent = objIds[0].GetObject(OpenMode.ForRead) as Entity;

            // ... и подсвечиваем вложенный примитив
            if (ent == null)
                return FullSubentityPath.Null;

            ent.Highlight(path, false);

            // Возвращаем путь к субпримитиву для возможности его затем погасить
            return path;
        }

        private static void UnhighlightSubEntities(FullSubentityPath[] paths)
        {
            foreach (FullSubentityPath path in paths)
            {
                if (path == FullSubentityPath.Null)
                    continue;

                ObjectId[] ids = path.GetObjectIds();

                Entity ent = ids[0].GetObject(OpenMode.ForRead) as Entity;

                if (ent != null)
                {
                    ent.Unhighlight(path, false);
                }
            }
        }

    }
}
