﻿// (C) Copyright 2002-20010 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// AcHelper is a utility class for the Associative Array Sample
//
// Products: AutoCAD (and verticals) 2012 and higher
//
// Written by Philippe Leefsma - 2010 - Developer Technical Services
// Перевод и адаптация под AutoCAD 2013+ - Александр Ривилис
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////
using System;

using Autodesk.AutoCAD.Runtime;
using Autodesk.AutoCAD.Geometry;
using Autodesk.AutoCAD.EditorInput;
using Autodesk.AutoCAD.DatabaseServices;
using Autodesk.AutoCAD.ApplicationServices;

[assembly: CommandClass(typeof(AssocArrayAPI.Commands))]

namespace AssocArrayAPI
{
  public class Commands : IExtensionApplication
  {
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Use: IExtensionApplication implementation
    //
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    public void Initialize()
    {
      Document doc = Application.DocumentManager.MdiActiveDocument;
      Editor ed = doc.Editor;

      ed.WriteMessage("\n --- Пример использования API для Ассоциативного массива --- ");

      ed.WriteMessage("\n Список команд: ");

      ed.WriteMessage("\n 1/  CreateAll");
      ed.WriteMessage("\n 2/  CreateAssocArrayRect");
      ed.WriteMessage("\n 3/  CreateAssocArrayRectSel");
      ed.WriteMessage("\n 4/  CreateAssocArrayPolar");
      ed.WriteMessage("\n 5/  CreateAssocArrayPath");
      ed.WriteMessage("\n 6/  AddSourceEntity");
      ed.WriteMessage("\n 7/  RemoveSourceEntity");
      ed.WriteMessage("\n 8/  ReplaceItems");
      ed.WriteMessage("\n 9/  RemoveItems");
      ed.WriteMessage("\n 10/ TransformItems");
      ed.WriteMessage("\n 11/ ResetItems");
      ed.WriteMessage("\n ------------------------------------------ ");
    }

    public void Terminate()
    {

    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Использование: Объединение всех команд по созданию массивов
    //
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    [CommandMethod("CreateAll")]
    public void CreateAll()
    {
      CreateAssocArrayRect();
      CreateAssocArrayPolar();
      CreateAssocArrayPath();

      Application.DocumentManager.MdiActiveDocument.SendStringToExecute("_ZOOM _E ", true, false, true);
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Использование: Создание Прямоугольного Ассоциативного Массива
    //
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    [CommandMethod("CreateAssocArrayRect")]
    public void CreateAssocArrayRect()
    {
      Document doc = Application.DocumentManager.MdiActiveDocument;
      Database db = doc.Database;
      Editor ed = doc.Editor;

      using (Transaction Tx = db.TransactionManager.StartTransaction()) {
        // Создаем исходный примитив - здесь простой круг (Circle)
        Circle circle = new Circle(new Point3d(0, 0, 0), new Vector3d(0, 0, 1), 1.0);
        circle.ColorIndex = 1;

        // Исходный примитив должен быть в базе
        AcHelper.PostToDatabase(circle);

        ObjectIdCollection sourceEntities = new ObjectIdCollection();

        sourceEntities.Add(circle.ObjectId);

        // Базовая точка массива в  VertexRef
        VertexRef basePoint = new VertexRef(new Point3d(0, 0, 0));

        // Параметры прямоугольного массива
        double columnSpacing = 3.0;
        double rowSpacing = 3.0;
        double levelSpacing = 2.0;
        int columnCount = 5;
        int rowCount = 5;
        int levelCount = 1;
        double rowElevation = 0;
        double axesAngle = 0;

        AssocArrayRectangularParameters parameters = new AssocArrayRectangularParameters(columnSpacing,
            rowSpacing,
            levelSpacing,
            columnCount,
            rowCount,
            levelCount,
            rowElevation,
            axesAngle);

        parameters.AxesAngle = Math.PI / 2;
        parameters.XAxisDirection = new Vector3d(1, 0, 0);
        parameters.YAxisDirection = new Vector3d(0, 1, 0);

        // Как показано ниже, параметры могут быть связаны с выражениями, включающими переменные
        AcHelper.AddOrModifVariable("RowCount", rowCount.ToString());
        AcHelper.AddOrModifVariable("RowSpacing", rowSpacing.ToString());

        parameters.SetRowCount(rowCount, "RowCount", "");
        parameters.SetRowSpacing(rowSpacing, "RowSpacing", "");

        // Создаем массив. Возвращается AssocArray помещенный в базу
        AssocArray array = Autodesk.AutoCAD.DatabaseServices.AssocArray.CreateArray(sourceEntities,
            basePoint,
            parameters);

        // Если нужно переместить массив, то имея его свойство EntityId
        // можно применить трансформацию как показано ниже
        Entity arrayEnt = Tx.GetObject(array.EntityId, OpenMode.ForWrite) as Entity;

        arrayEnt.TransformBy(Matrix3d.Displacement(new Vector3d(5, 5, 5)));

        // Заставляем обновится все связи
        AssocManager.EvaluateTopLevelNetwork(db, null, 0);

        // Так как массив уже создан, то исходный круг можно удалить.
        // Он уже содержится в массиве (фактически в анонимном блоке)
        circle.Erase(true);

        Tx.Commit();
      }
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Использование: Создание прямоугольного массива на основе выбранных примитивов
    //
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    [CommandMethod("CreateAssocArrayRectSel")]
    public void CreateAssocArrayRectSel()
    {
      Document doc = Application.DocumentManager.MdiActiveDocument;
      Database db = doc.Database;
      Editor ed = doc.Editor;

      PromptSelectionOptions pso = new PromptSelectionOptions();

      PromptSelectionResult psr = ed.GetSelection(pso);

      if (psr.Status != PromptStatus.OK)
        return;

      PromptPointResult ppr = ed.GetPoint("\nУкажите базовую точку: ");

      if (ppr.Status != PromptStatus.OK)
        return;

      using (Transaction Tx = db.TransactionManager.StartTransaction()) {
        ObjectIdCollection sourceEntities = new ObjectIdCollection(psr.Value.GetObjectIds());

        VertexRef basePoint = new VertexRef(ppr.Value);

        double columnSpacing = 5.0;
        double rowSpacing = 5.0;
        double levelSpacing = 2.0;
        int columnCount = 5;
        int rowCount = 5;
        int levelCount = 3;
        double rowElevation = 0;
        double axesAngle = 0;

        AssocArrayRectangularParameters parameters = new AssocArrayRectangularParameters(columnSpacing,
            rowSpacing,
            levelSpacing,
            columnCount,
            rowCount,
            levelCount,
            rowElevation,
            axesAngle);

        parameters.AxesAngle = Math.PI / 2;
        parameters.XAxisDirection = new Vector3d(1, 0, 0);
        parameters.YAxisDirection = new Vector3d(0, 1, 0);

        AssocArray array = Autodesk.AutoCAD.DatabaseServices.AssocArray.CreateArray(sourceEntities,
            basePoint,
            parameters);

        AssocManager.EvaluateTopLevelNetwork(db, null, 0);

        Tx.Commit();
      }
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Использование: Создание кругового массива
    //
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    [CommandMethod("CreateAssocArrayPolar")]
    public void CreateAssocArrayPolar()
    {
      Document doc = Application.DocumentManager.MdiActiveDocument;
      Database db = doc.Database;
      Editor ed = doc.Editor;

      using (Transaction Tx = db.TransactionManager.StartTransaction()) {
        // Добавим профили в базу - это будут круг и отрезок
        Circle circle = new Circle(new Point3d(35, 10, 0), new Vector3d(0, 0, 1), 1.0);
        circle.ColorIndex = 2;

        AcHelper.PostToDatabase(circle);

        Line line = new Line(new Point3d(34, 9, 0), new Point3d(36, 11, 0));
        line.ColorIndex = 2;

        AcHelper.PostToDatabase(line);

        ObjectIdCollection sourceEntities = new ObjectIdCollection();

        sourceEntities.Add(line.ObjectId);
        sourceEntities.Add(circle.ObjectId);

        VertexRef basePoint = new VertexRef(circle.Center);

        // Создаем и инициализируем параметры кругового массива
        AssocArrayPolarParameters parameters = new AssocArrayPolarParameters();

        parameters.AngleBetweenItems = 10.0;
        parameters.Direction = AssocArrayPolarParameters.ArcDirection.CounterClockwise;
        parameters.FillAngle = 180.0;
        parameters.ItemCount = 5;
        // parameters.Method = AssocArrayPolarParameters.MethodType.NumberAndFillAngle; // Этого метода в AutoCAD 2013+ нет
        parameters.Radius = 4.0;
        parameters.RotateItems = true;
        parameters.StartAngle = 0.0;

        // Создаем массив
        AssocArray array = Autodesk.AutoCAD.DatabaseServices.AssocArray.CreateArray(sourceEntities,
            basePoint,
            parameters);

        // Перемещаем массив
        Entity arrayEnt = Tx.GetObject(array.EntityId, OpenMode.ForWrite) as Entity;

        arrayEnt.TransformBy(Matrix3d.Displacement(circle.Center.GetAsVector()));

        AssocManager.EvaluateTopLevelNetwork(db, null, 0);

        Tx.Commit();
      }
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Использование: Создание массива по траектории (траектория - сплайн)
    //
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    [CommandMethod("CreateAssocArrayPath")]
    public void CreateAssocArrayPath()
    {
      Document doc = Application.DocumentManager.MdiActiveDocument;
      Database db = doc.Database;
      Editor ed = doc.Editor;

      using (Transaction Tx = db.TransactionManager.StartTransaction()) {
        // Добавим профили в базу - это будут круг и отрезок
        Circle circle = new Circle(new Point3d(-10, 0, 0), new Vector3d(0, 0, 1), 1.0);
        circle.ColorIndex = 3;

        AcHelper.PostToDatabase(circle);

        Line line = new Line(new Point3d(-11, -1, 0), new Point3d(-9, 1, 0));
        line.ColorIndex = 3;

        AcHelper.PostToDatabase(line);

        // Добавляем примитив траектории (сплайн) в базу

        double fitTolerance = 0.0;
        int order = 4;
        int ptsCount = 5;

        double[] pts = new double[]{-6.8447, 0.9430,  0,
                                            -5.1409, -3.0562, 0,
                                            -2.4095, 2.1049,  0,
                                             3.4590, -3.9479, 0,
                                             7.1370, 6.3472,  0};

        Point3dCollection ptsColl = new Point3dCollection();

        for (int i = 0, j = 0; i < ptsCount; i++) {
          ptsColl.Add(new Point3d(pts[j++], pts[j++], pts[j++]));
        }

        Spline spline = new Spline(ptsColl, order, fitTolerance);

        AcHelper.PostToDatabase(spline);

        Matrix3d offset = Matrix3d.Displacement(new Vector3d(60, 10, 0));

        circle.TransformBy(offset);
        line.TransformBy(offset);
        spline.TransformBy(offset);

        ObjectIdCollection sourceEntities = new ObjectIdCollection();

        sourceEntities.Add(circle.ObjectId);
        sourceEntities.Add(line.ObjectId);

        VertexRef basePoint = new VertexRef(circle.Center);

        int itemCount = 6;
        double itemSpacing = spline.GetDistanceAtParameter(spline.EndParam) / itemCount;
        double rowSpacing = 1.0;
        double levelSpacing = 1.0;
        int rowCount = 0;
        int levelCount = 0;
        double rowElevation = 5.0;

        AssocArrayPathParameters parameters = new AssocArrayPathParameters(itemSpacing,
            rowSpacing,
            levelSpacing,
            itemCount,
            rowCount,
            levelCount,
            rowElevation);

        parameters.Method = AssocArrayPathParameters.MethodType.Measure;

        parameters.Path = new EdgeRef(spline);

        //parameters.AlignItems = true;
        //parameters.EndOffset = 0.0;
        //parameters.ItemCount = 5;
        //parameters.ItemSpacing = 1.0;
        //parameters.PathDirection = true;
        //parameters.StartOffset = 0.0;

        AssocArray array = Autodesk.AutoCAD.DatabaseServices.AssocArray.CreateArray(sourceEntities,
            basePoint,
            parameters);

        AssocManager.EvaluateTopLevelNetwork(db, null, 0);

        Tx.Commit();
      }
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Использование: Добавляем исходный примитив в выбранный массив
    //
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    [CommandMethod("AddSourceEntity")]
    public void AddSourceEntity()
    {
      Document doc = Application.DocumentManager.MdiActiveDocument;
      Database db = doc.Database;
      Editor ed = doc.Editor;

      PromptEntityOptions peo1 = new PromptEntityOptions("\nВыберите ассоциативный массив: ");

      PromptEntityResult per1 = ed.GetEntity(peo1);

      if (per1.Status != PromptStatus.OK)
        return;

      if (!AssocArray.IsAssociativeArray(per1.ObjectId)) {
        ed.WriteMessage("\nЭто не ассоциативный массив, повторите...");
        return;
      }

      PromptEntityOptions peo2 = new PromptEntityOptions("\nВыберите исходный примитив для добавления: ");

      PromptEntityResult per2 = ed.GetEntity(peo2);

      if (per2.Status != PromptStatus.OK)
        return;

      PromptPointResult ppr = ed.GetPoint("\nУкажите базовую точку: ");

      if (ppr.Status != PromptStatus.OK)
        return;

      using (Transaction Tx = db.TransactionManager.StartTransaction()) {
        AssocArray array = AssocArray.GetAssociativeArray(per1.ObjectId);

        array.AddSourceEntity(per2.ObjectId, ppr.Value);

        AssocManager.EvaluateTopLevelNetwork(db, null, 0);

        Tx.Commit();
      }
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Использование: Удаление исходного примитива из массива. 
    //      Только для простоты удалим первый элемент из массива
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    [CommandMethod("RemoveSourceEntity")]
    public void RemoveSourceEntity()
    {
      Document doc = Application.DocumentManager.MdiActiveDocument;
      Database db = doc.Database;
      Editor ed = doc.Editor;

      PromptEntityOptions peo1 = new PromptEntityOptions("\nВыберите ассоциативный массив: ");

      PromptEntityResult per1 = ed.GetEntity(peo1);

      if (per1.Status != PromptStatus.OK)
        return;

      if (!AssocArray.IsAssociativeArray(per1.ObjectId)) {
        ed.WriteMessage("\nЭто не ассоциативный массив, повторите...");
        return;
      }

      using (Transaction Tx = db.TransactionManager.StartTransaction()) {
        AssocArray array = AssocArray.GetAssociativeArray(per1.ObjectId);

        if (array.SourceEntities.Count < 2) {
          ed.WriteMessage("\nВыбранный массив должен содержать как минимуму два исходных примитива, повторите...");
          return;
        }

        array.RemoveSourceEntity(array.SourceEntities[0]);

        AssocManager.EvaluateTopLevelNetwork(db, null, 0);

        Tx.Commit();
      }
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Использование: Замена выбранных примитивов массива исходным примитивом
    //
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    [CommandMethod("ReplaceItems")]
    public void ReplaceItems()
    {
      Document doc = Application.DocumentManager.MdiActiveDocument;
      Database db = doc.Database;
      Editor ed = doc.Editor;

      PromptEntityOptions peo1 = new PromptEntityOptions("\nВыберите ассоциативный массив: ");

      PromptEntityResult per1 = ed.GetEntity(peo1);

      if (per1.Status != PromptStatus.OK)
        return;

      if (!AssocArray.IsAssociativeArray(per1.ObjectId)) {
        ed.WriteMessage("\nЭто не ассоциативный массив, повторите...");
        return;
      }

      FullSubentityPath[] paths;

      if (AcHelper.SelectNestedEntities("\nВыберите элементы массива: ", out paths) != PromptStatus.OK)
        return;

      PromptEntityOptions peo2 = new PromptEntityOptions("\nВыберите примитив для замены: ");

      PromptEntityResult per2 = ed.GetEntity(peo2);

      if (per2.Status != PromptStatus.OK)
        return;

      PromptPointResult ppr = ed.GetPoint("\nУкажите базовую точку: ");

      if (ppr.Status != PromptStatus.OK)
        return;

      using (Transaction Tx = db.TransactionManager.StartTransaction()) {
        AssocArray array = AssocArray.GetAssociativeArray(per1.ObjectId);

        ItemLocator[] ItemLocators = AssocArray.getItemLocators(paths);

        ObjectIdCollection substEntities = new ObjectIdCollection();

        substEntities.Add(per2.ObjectId);

        VertexRef basePoint = new VertexRef(ppr.Value);

        array.ReplaceItems(ItemLocators, substEntities, basePoint);

        AssocManager.EvaluateTopLevelNetwork(db, null, 0);

        Tx.Commit();
      }
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Использование: Удаление выбранных элементов из массива
    //
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    [CommandMethod("RemoveItems")]
    public void RemoveItems()
    {
      Document doc = Application.DocumentManager.MdiActiveDocument;
      Database db = doc.Database;
      Editor ed = doc.Editor;

      PromptEntityOptions peo1 = new PromptEntityOptions("\nВыберите ассоциативный массив: ");

      PromptEntityResult per1 = ed.GetEntity(peo1);

      if (per1.Status != PromptStatus.OK)
        return;

      if (!AssocArray.IsAssociativeArray(per1.ObjectId)) {
        ed.WriteMessage("\nЭто не ассоциативный массив, повторите...");
        return;
      }

      FullSubentityPath[] paths;

      if (AcHelper.SelectNestedEntities("\nВыберите элементы массива: ", out paths) != PromptStatus.OK)
        return;

      using (Transaction Tx = db.TransactionManager.StartTransaction()) {
        AssocArray array = AssocArray.GetAssociativeArray(per1.ObjectId);

        ItemLocator[] ItemLocators = AssocArray.getItemLocators(paths);

        foreach (ItemLocator itemLocator in ItemLocators) {
          array.DeleteItem(itemLocator, true);
        }

        AssocManager.EvaluateTopLevelNetwork(db, null, 0);

        Tx.Commit();
      }
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Use: Transform selected array items using the translation provided by the user
    //
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    [CommandMethod("TransformItems")]
    public void TransformItems()
    {
      Document doc = Application.DocumentManager.MdiActiveDocument;
      Database db = doc.Database;
      Editor ed = doc.Editor;

      PromptEntityOptions peo1 = new PromptEntityOptions("\nВыберите ассоциативный массив: ");

      PromptEntityResult per1 = ed.GetEntity(peo1);

      if (per1.Status != PromptStatus.OK)
        return;

      if (!AssocArray.IsAssociativeArray(per1.ObjectId)) {
        ed.WriteMessage("\nЭто не ассоциативный массив, повторите...");
        return;
      }

      FullSubentityPath[] paths;

      if (AcHelper.SelectNestedEntities("\nВыберите элементы массива: ", out paths) != PromptStatus.OK)
        return;

      PromptPointOptions ppo = new PromptPointOptions("\nУкажите смещение: ");

      PromptPointResult ppr1 = ed.GetPoint(ppo);

      if (ppr1.Status != PromptStatus.OK)
        return;

      ppo.UseBasePoint = true;
      ppo.BasePoint = ppr1.Value;

      PromptPointResult ppr2 = ed.GetPoint(ppo);

      if (ppr2.Status != PromptStatus.OK)
        return;

      using (Transaction Tx = db.TransactionManager.StartTransaction()) {
        AssocArray array = AssocArray.GetAssociativeArray(per1.ObjectId);

        Matrix3d transform = Matrix3d.Displacement(ppr2.Value.GetAsVector().Subtract(ppr1.Value.GetAsVector()));

        ItemLocator[] ItemLocators = AssocArray.getItemLocators(paths);

        foreach (ItemLocator itemLocator in ItemLocators) {
          array.TransformItemBy(itemLocator, transform);
        }

        AssocManager.EvaluateTopLevelNetwork(db, null, 0);

        Tx.Commit();
      }
    }

    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    // Использование: Сбросим массив до оригинальной конфигурации, т.е. сброс всех переопределений
    //
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    [CommandMethod("ResetItems")]
    public void ResetItems()
    {
      Document doc = Application.DocumentManager.MdiActiveDocument;
      Database db = doc.Database;
      Editor ed = doc.Editor;

      PromptEntityOptions peo1 = new PromptEntityOptions("\nВыберите ассоциативный массив: ");

      PromptEntityResult per1 = ed.GetEntity(peo1);

      if (per1.Status != PromptStatus.OK)
        return;

      if (!AssocArray.IsAssociativeArray(per1.ObjectId)) {
        ed.WriteMessage("\nЭто не ассоциативный массив, повторите...");
        return;
      }

      using (Transaction Tx = db.TransactionManager.StartTransaction()) {
        AssocArray array = AssocArray.GetAssociativeArray(per1.ObjectId);

        array.ResetItems();

        AssocManager.EvaluateTopLevelNetwork(db, null, 0);

        Tx.Commit();
      }
    }
  }
}
